@@ -36,7 +36,6 @@
 #include <jsapi.h>
 
 #include <girepository.h>
-#include <girffi.h>
 #include <sys/mman.h>
 #include <unistd.h>
 #include <errno.h>
@@ -47,17 +46,10 @@
  */
 #define GJS_ARG_INDEX_INVALID G_MAXUINT8
 
-typedef enum {
-    PARAM_NORMAL,
-    PARAM_SKIPPED,
-    PARAM_ARRAY,
-    PARAM_CALLBACK
-} ParamType;
-
 typedef struct {
     GIFunctionInfo *info;
 
-    ParamType *param_types;
+    GjsParamType *param_types;
 
     guint8 expected_js_argc;
     guint8 js_out_argc;
@@ -72,16 +64,6 @@ static struct JSClass gjs_function_class;
  */
 static GSList *completed_trampolines = NULL;  /* GjsCallbackTrampoline */
 
-typedef struct {
-    gint ref_count;
-    JSRuntime *runtime;
-    GICallableInfo *info;
-    jsval js_function;
-    ffi_cif cif;
-    ffi_closure *closure;
-    GIScopeType scope;
-} GjsCallbackTrampoline;
-
 GJS_DEFINE_PRIV_FROM_JS(Function, gjs_function_class)
 
 /*
@@ -123,13 +105,13 @@ function_new_resolve(JSContext *context,
     return JS_TRUE;
 }
 
-static void
+void
 gjs_callback_trampoline_ref(GjsCallbackTrampoline *trampoline)
 {
     trampoline->ref_count++;
 }
 
-static void
+void
 gjs_callback_trampoline_unref(GjsCallbackTrampoline *trampoline)
 {
     /* Not MT-safe, like all the rest of GJS */
@@ -140,9 +122,11 @@ gjs_callback_trampoline_unref(GjsCallbackTrampoline *trampoline)
 
         context = gjs_runtime_get_current_context(trampoline->runtime);
 
-        JS_RemoveValueRoot(context, &trampoline->js_function);
+        if (!trampoline->is_vfunc)
+            JS_RemoveValueRoot(context, &trampoline->js_function);
         g_callable_info_free_closure(trampoline->info, trampoline->closure);
         g_base_info_unref( (GIBaseInfo*) trampoline->info);
+        g_free (trampoline->param_types);
         g_slice_free(GjsCallbackTrampoline, trampoline);
     }
 }
@@ -197,6 +181,8 @@ set_return_ffi_arg_from_giargument (GITypeInfo  *ret_type,
                 *(ffi_arg *) result = (ffi_arg) return_value->v_pointer;
                 break;
             }
+
+            g_base_info_unref(interface_info);
         }
     default:
         *(ffi_arg *) result = (ffi_arg) return_value->v_uint64;
@@ -219,11 +205,12 @@ gjs_callback_closure(ffi_cif *cif,
 {
     JSContext *context;
     GjsCallbackTrampoline *trampoline;
-    int i, n_args, n_jsargs;
+    int i, n_args, n_jsargs, n_outargs;
     jsval *jsargs, rval;
+    JSObject *this_object;
     GITypeInfo ret_type;
-    GArgument return_value;
     gboolean success = FALSE;
+    gboolean ret_type_is_void;
 
     trampoline = data;
     g_assert(trampoline);
@@ -235,10 +222,12 @@ gjs_callback_closure(ffi_cif *cif,
 
     g_assert(n_args >= 0);
 
+    n_outargs = 0;
     jsargs = (jsval*)g_newa(jsval, n_args);
     for (i = 0, n_jsargs = 0; i < n_args; i++) {
         GIArgInfo arg_info;
         GITypeInfo type_info;
+        GjsParamType param_type;
 
         g_callable_info_load_arg(trampoline->info, i, &arg_info);
         g_arg_info_load_type(&arg_info, &type_info);
@@ -247,15 +236,56 @@ gjs_callback_closure(ffi_cif *cif,
         if (g_type_info_get_tag(&type_info) == GI_TYPE_TAG_VOID)
             continue;
 
-        if (!gjs_value_from_g_argument(context,
-                                       &jsargs[n_jsargs++],
-                                       &type_info,
-                                       args[i]))
-            goto out;
+        if (g_arg_info_get_direction(&arg_info) == GI_DIRECTION_OUT) {
+            n_outargs++;
+            continue;
+        }
+
+        if (g_arg_info_get_direction(&arg_info) == GI_DIRECTION_INOUT)
+            n_outargs++;
+
+        param_type = trampoline->param_types[i];
+
+        switch (param_type) {
+            case PARAM_SKIPPED:
+                continue;
+            case PARAM_ARRAY: {
+                gint array_length_pos = g_type_info_get_array_length(&type_info);
+                GIArgInfo array_length_arg;
+                GITypeInfo arg_type_info;
+                jsval length;
+
+                g_callable_info_load_arg(trampoline->info, array_length_pos, &array_length_arg);
+                g_arg_info_load_type(&array_length_arg, &arg_type_info);
+                if (!gjs_value_from_g_argument(context, &length,
+                                               &arg_type_info,
+                                               args[array_length_pos], TRUE))
+                    goto out;
+
+                if (!gjs_value_from_explicit_array(context, &jsargs[n_jsargs++],
+                                                   &type_info, args[i], JSVAL_TO_INT(length)))
+                    goto out;
+                break;
+            }
+            case PARAM_NORMAL:
+                if (!gjs_value_from_g_argument(context,
+                                               &jsargs[n_jsargs++],
+                                               &type_info,
+                                               args[i], FALSE))
+                    goto out;
+                break;
+            default:
+                g_assert_not_reached();
+        }
     }
 
+    if (trampoline->is_vfunc)
+        this_object = JSVAL_TO_OBJECT(jsargs[0]);
+    else
+        this_object = NULL;
+
     if (!JS_CallFunctionValue(context,
-                              NULL,
+                              this_object,
                               trampoline->js_function,
                               n_jsargs,
                               jsargs,
@@ -264,21 +294,103 @@ gjs_callback_closure(ffi_cif *cif,
     }
 
     g_callable_info_load_return_type(trampoline->info, &ret_type);
+    ret_type_is_void = g_type_info_get_tag (&ret_type) == GI_TYPE_TAG_VOID;
 
-    if (!gjs_value_to_g_argument(context,
-                                 rval,
-                                 &ret_type,
-                                 "callback",
-                                 GJS_ARGUMENT_RETURN_VALUE,
-                                 FALSE,
-                                 TRUE,
-                                 &return_value)) {
-        goto out;
+    if (n_outargs == 0 && !ret_type_is_void) {
+        GIArgument argument;
+
+        /* non-void return value, no out args. Should
+         * be a single return value. */
+        if (!gjs_value_to_g_argument(context,
+                                     rval,
+                                     &ret_type,
+                                     "callback",
+                                     GJS_ARGUMENT_RETURN_VALUE,
+                                     GI_TRANSFER_NOTHING,
+                                     TRUE,
+                                     &argument))
+            goto out;
+
+        set_return_ffi_arg_from_giargument(&ret_type,
+                                           result,
+                                           &argument);
+    } else if (n_outargs == 1 && ret_type_is_void) {
+        /* void return value, one out args. Should
+         * be a single return value. */
+        for (i = 0; i < n_args; i++) {
+            GIArgInfo arg_info;
+            GITypeInfo type_info;
+            g_callable_info_load_arg(trampoline->info, i, &arg_info);
+            if (g_arg_info_get_direction(&arg_info) == GI_DIRECTION_IN)
+                continue;
+
+            g_arg_info_load_type(&arg_info, &type_info);
+            if (!gjs_value_to_g_argument(context,
+                                         rval,
+                                         &type_info,
+                                         "callback",
+                                         GJS_ARGUMENT_ARGUMENT,
+                                         GI_TRANSFER_NOTHING,
+                                         TRUE,
+                                         *(gpointer *)args[i]))
+                goto out;
+
+            break;
+        }
+    } else {
+        jsval elem;
+        gsize elem_idx = 0;
+        /* more than one of a return value or an out argument.
+         * Should be an array of output values. */
+
+        if (!ret_type_is_void) {
+            GIArgument argument;
+
+            if (!JS_GetElement(context, JSVAL_TO_OBJECT(rval), elem_idx, &elem))
+                goto out;
+
+            if (!gjs_value_to_g_argument(context,
+                                         elem,
+                                         &ret_type,
+                                         "callback",
+                                         GJS_ARGUMENT_ARGUMENT,
+                                         GI_TRANSFER_NOTHING,
+                                         TRUE,
+                                         &argument))
+                goto out;
+
+            set_return_ffi_arg_from_giargument(&ret_type,
+                                               result,
+                                               &argument);
+
+            elem_idx++;
+        }
+
+        for (i = 0; i < n_args; i++) {
+            GIArgInfo arg_info;
+            GITypeInfo type_info;
+            g_callable_info_load_arg(trampoline->info, i, &arg_info);
+            if (g_arg_info_get_direction(&arg_info) == GI_DIRECTION_IN)
+                continue;
+
+            g_arg_info_load_type(&arg_info, &type_info);
+            if (!JS_GetElement(context, JSVAL_TO_OBJECT(rval), elem_idx, &elem))
+                goto out;
+
+            if (!gjs_value_to_g_argument(context,
+                                         elem,
+                                         &type_info,
+                                         "callback",
+                                         GJS_ARGUMENT_ARGUMENT,
+                                         GI_TRANSFER_NOTHING,
+                                         TRUE,
+                                         *(gpointer *)args[i]))
+                goto out;
+
+            elem_idx++;
+        }
     }
 
-    
-    set_return_ffi_arg_from_giargument(&ret_type, result, &return_value);
-
     success = TRUE;
 
 out:
@@ -309,15 +421,17 @@ gjs_destroy_notify_callback(gpointer data)
     gjs_callback_trampoline_unref(trampoline);
 }
 
-static GjsCallbackTrampoline*
+GjsCallbackTrampoline*
 gjs_callback_trampoline_new(JSContext      *context,
                             jsval           function,
                             GICallableInfo *callable_info,
-                            GIScopeType     scope)
+                            GIScopeType     scope,
+                            gboolean        is_vfunc)
 {
     GjsCallbackTrampoline *trampoline;
+    int n_args, i;
 
-    if (function == JSVAL_NULL) {
+    if (JSVAL_IS_NULL(function)) {
         return NULL;
     }
 
@@ -329,11 +443,70 @@ gjs_callback_trampoline_new(JSContext      *context,
     trampoline->info = callable_info;
     g_base_info_ref((GIBaseInfo*)trampoline->info);
     trampoline->js_function = function;
-    JS_AddValueRoot(context, &trampoline->js_function);
+    if (!is_vfunc)
+        JS_AddValueRoot(context, &trampoline->js_function);
+
+    /* Analyze param types and directions, similarly to init_cached_function_data */
+    n_args = g_callable_info_get_n_args(trampoline->info);
+    trampoline->param_types = g_new0(GjsParamType, n_args);
+
+    for (i = 0; i < n_args; i++) {
+        GIDirection direction;
+        GIArgInfo arg_info;
+        GITypeInfo type_info;
+        GITypeTag type_tag;
+
+        if (trampoline->param_types[i] == PARAM_SKIPPED)
+            continue;
+
+        g_callable_info_load_arg(trampoline->info, i, &arg_info);
+        g_arg_info_load_type(&arg_info, &type_info);
+
+        direction = g_arg_info_get_direction(&arg_info);
+        type_tag = g_type_info_get_tag(&type_info);
+
+        if (direction != GI_DIRECTION_IN) {
+            /* INOUT and OUT arguments are handled differently. */
+            continue;
+        }
+
+        if (type_tag == GI_TYPE_TAG_INTERFACE) {
+            GIBaseInfo* interface_info;
+            GIInfoType interface_type;
+
+            interface_info = g_type_info_get_interface(&type_info);
+            interface_type = g_base_info_get_type(interface_info);
+            if (interface_type == GI_INFO_TYPE_CALLBACK) {
+                gjs_throw(context, "Callback accepts another callback as a parameter. This is not supported");
+                g_base_info_unref(interface_info);
+                return NULL;
+            }
+            g_base_info_unref(interface_info);
+        } else if (type_tag == GI_TYPE_TAG_ARRAY) {
+            if (g_type_info_get_array_type(&type_info) == GI_ARRAY_TYPE_C) {
+                int array_length_pos = g_type_info_get_array_length(&type_info);
+
+                if (array_length_pos >= 0 && array_length_pos < n_args) {
+                    GIArgInfo length_arg_info;
+
+                    g_callable_info_load_arg(trampoline->info, array_length_pos, &length_arg_info);
+                    if (g_arg_info_get_direction(&length_arg_info) != direction) {
+                        gjs_throw(context, "Callback has an array with different-direction length arg, not supported");
+                        return NULL;
+                    }
+
+                    trampoline->param_types[array_length_pos] = PARAM_SKIPPED;
+                    trampoline->param_types[i] = PARAM_ARRAY;
+                }
+            }
+        }
+    }
+
     trampoline->closure = g_callable_info_prepare_closure(callable_info, &trampoline->cif,
                                                           gjs_callback_closure, trampoline);
 
     trampoline->scope = scope;
+    trampoline->is_vfunc = is_vfunc;
 
     return trampoline;
 }
@@ -436,6 +609,8 @@ set_gargument_from_ffi_return_value (GITypeInfo                  *return_info,
                 return_value->v_pointer = (gpointer) value->rv_ffi_arg;
                 break;
             }
+
+            g_base_info_unref(interface_info);
         }
         break;
     default:
@@ -481,10 +656,10 @@ gjs_invoke_c_function(JSContext      *context,
     GError *local_error = NULL;
     gboolean failed, postinvoke_release_failed;
 
-    GIFunctionInfoFlags flags;
     gboolean is_method;
     GITypeInfo return_info;
     GITypeTag return_tag;
+    GIInfoType info_type;
     jsval *return_values = NULL;
     guint8 next_rval = 0; /* index into return_values */
     GSList *iter;
@@ -502,9 +677,33 @@ gjs_invoke_c_function(JSContext      *context,
         completed_trampolines = NULL;
     }
 
-    flags = g_function_info_get_flags(function->info);
-    is_method = (flags & GI_FUNCTION_IS_METHOD) != 0;
-    can_throw_gerror = (flags & GI_FUNCTION_THROWS) != 0;
+    info_type = g_base_info_get_type((GIBaseInfo *)function->info);
+
+    switch (info_type) {
+    case GI_INFO_TYPE_FUNCTION:
+        {
+            GIFunctionInfoFlags flags;
+            flags = g_function_info_get_flags((GIFunctionInfo *)function->info);
+            is_method = (flags & GI_FUNCTION_IS_METHOD) != 0;
+            can_throw_gerror = (flags & GI_FUNCTION_THROWS) != 0;
+        }
+        break;
+    case GI_INFO_TYPE_VFUNC:
+        {
+            GIVFuncInfoFlags flags;
+            flags = g_vfunc_info_get_flags ((GIVFuncInfo *)function->info);
+            can_throw_gerror = (flags & GI_VFUNC_THROWS) != 0;
+        }
+        is_method = TRUE;
+      break;
+    case GI_INFO_TYPE_CALLBACK:
+        is_method = TRUE;
+        can_throw_gerror = FALSE;
+        break;
+    default:
+        g_assert_not_reached();
+    }
+
     c_argc = function->invoker.cif.nargs;
     gi_argc = g_callable_info_get_n_args( (GICallableInfo*) function->info);
 
@@ -558,6 +757,13 @@ gjs_invoke_c_function(JSContext      *context,
             GType gtype;
 
             in_arg_cvalues[0].v_pointer = gjs_g_object_from_object(context, obj);
+            if (in_arg_cvalues[0].v_pointer == NULL) {
+                /* priv == NULL (user probably forgot to chain _init).
+                 * Anyway, in this case we've thrown an exception, so just
+                 * make sure we fail. */
+                failed = TRUE;
+                goto release;
+            }
 
             gtype = g_registered_type_info_get_g_type ((GIRegisteredTypeInfo *)container);
             if (!g_type_is_a (G_TYPE_FROM_INSTANCE (in_arg_cvalues[0].v_pointer),
@@ -636,7 +842,7 @@ gjs_invoke_c_function(JSContext      *context,
         } else {
             GArgument *in_value;
             GITypeInfo ainfo;
-            ParamType param_type;
+            GjsParamType param_type;
 
             g_arg_info_load_type(&arg_info, &ainfo);
 
@@ -669,7 +875,8 @@ gjs_invoke_c_function(JSContext      *context,
                     trampoline = gjs_callback_trampoline_new(context,
                                                              value,
                                                              callable_info,
-                                                             scope);
+                                                             scope,
+                                                             FALSE);
                     closure = trampoline->closure;
                     g_base_info_unref(callable_info);
                 }
@@ -837,7 +1044,8 @@ gjs_invoke_c_function(JSContext      *context,
                 array_length_pos += is_method ? 1 : 0;
                 arg_failed = !gjs_value_from_g_argument(context, &length,
                                                         &arg_type_info,
-                                                        &out_arg_cvalues[array_length_pos]);
+                                                        &out_arg_cvalues[array_length_pos],
+                                                        TRUE);
                 if (!arg_failed) {
                     arg_failed = !gjs_value_from_explicit_array(context,
                                                                 &return_values[next_rval],
@@ -854,7 +1062,8 @@ gjs_invoke_c_function(JSContext      *context,
                     failed = TRUE;
             } else {
                 arg_failed = !gjs_value_from_g_argument(context, &return_values[next_rval],
-                                                        &return_info, &return_gargument);
+                                                        &return_info, &return_gargument,
+                                                        TRUE);
                 /* Free GArgument, the jsval should have ref'd or copied it */
                 if (!arg_failed &&
                     !gjs_g_argument_release(context,
@@ -880,7 +1089,7 @@ release:
         GIDirection direction;
         GIArgInfo arg_info;
         GITypeInfo arg_type_info;
-        ParamType param_type;
+        GjsParamType param_type;
 
         g_callable_info_load_arg( (GICallableInfo*) function->info, gi_arg_pos, &arg_info);
         direction = g_arg_info_get_direction(&arg_info);
@@ -974,7 +1183,8 @@ release:
                 array_length_pos += is_method ? 1 : 0;
                 arg_failed = !gjs_value_from_g_argument(context, &array_length,
                                                         &array_length_type_info,
-                                                        &out_arg_cvalues[array_length_pos]);
+                                                        &out_arg_cvalues[array_length_pos],
+                                                        TRUE);
                 if (!arg_failed) {
                     arg_failed = !gjs_value_from_explicit_array(context,
                                                                 &return_values[next_rval],
@@ -986,7 +1196,8 @@ release:
                 arg_failed = !gjs_value_from_g_argument(context,
                                                         &return_values[next_rval],
                                                         &arg_type_info,
-                                                        arg);
+                                                        arg,
+                                                        TRUE);
             }
 
             if (arg_failed)
@@ -1191,10 +1402,14 @@ function_to_string (JSContext *context,
     if (priv == NULL) {
         string = "function () {\n}";
         free = FALSE;
-    } else {
+    } else if (g_base_info_get_type(priv->info) == GI_INFO_TYPE_FUNCTION) {
         string = g_strdup_printf("function %s(){\n\t/* proxy for native symbol %s(); */\n}",
                                  g_base_info_get_name ((GIBaseInfo *) priv->info),
-                                 g_function_info_get_symbol (priv->info));
+                                 g_function_info_get_symbol ((GIFunctionInfo *) priv->info));
+        free = TRUE;
+    } else {
+        string = g_strdup_printf("function %s(){\n\t/* proxy for native symbol */\n}",
+                                 g_base_info_get_name ((GIBaseInfo *) priv->info));
         free = TRUE;
     }
 
@@ -1250,15 +1465,40 @@ static JSFunctionSpec gjs_function_proto_funcs[] = {
 static gboolean
 init_cached_function_data (JSContext      *context,
                            Function       *function,
-                           GIFunctionInfo *info)
+                           GType           gtype,
+                           GICallableInfo *info)
 {
     guint8 i, n_args, array_length_pos;
     GError *error = NULL;
     GITypeInfo return_type;
+    GIInfoType info_type;
 
-    if (!g_function_info_prep_invoker(info, &(function->invoker), &error)) {
-        gjs_throw_g_error(context, error);
-        return FALSE;
+    info_type = g_base_info_get_type((GIBaseInfo *)info);
+
+    if (info_type == GI_INFO_TYPE_FUNCTION) {
+        if (!g_function_info_prep_invoker((GIFunctionInfo *)info,
+                                          &(function->invoker),
+                                          &error)) {
+            gjs_throw_g_error(context, error);
+            return FALSE;
+        }
+    } else if (info_type == GI_INFO_TYPE_VFUNC) {
+        gpointer addr;
+
+        addr = g_vfunc_info_get_address((GIVFuncInfo *)info, gtype, &error);
+        if (error != NULL) {
+            if (error->code != G_INVOKE_ERROR_SYMBOL_NOT_FOUND)
+                gjs_throw_g_error(context, error);
+
+            return FALSE;
+        }
+
+        if (!g_function_invoker_new_for_address(addr, info,
+                                                &(function->invoker),
+                                                &error)) {
+            gjs_throw_g_error(context, error);
+            return FALSE;
+        }
     }
 
     g_callable_info_load_return_type((GICallableInfo*)info, &return_type);
@@ -1266,7 +1506,7 @@ init_cached_function_data (JSContext      *context,
         function->js_out_argc += 1;
 
     n_args = g_callable_info_get_n_args((GICallableInfo*) info);
-    function->param_types = g_new0(ParamType, n_args);
+    function->param_types = g_new0(GjsParamType, n_args);
 
     array_length_pos = g_type_info_get_array_length(&return_type);
     if (array_length_pos >= 0 && array_length_pos < n_args)
@@ -1317,6 +1557,7 @@ init_cached_function_data (JSContext      *context,
                         gjs_throw(context, "Function %s.%s has a GDestroyNotify but no user_data, not supported",
                                   g_base_info_get_namespace( (GIBaseInfo*) info),
                                   g_base_info_get_name( (GIBaseInfo*) info));
+                        g_base_info_unref(interface_info);
                         return JS_FALSE;
                     }
                 }
@@ -1368,7 +1609,8 @@ init_cached_function_data (JSContext      *context,
 
 static JSObject*
 function_new(JSContext      *context,
-             GIFunctionInfo *info)
+             GType           gtype,
+             GICallableInfo *info)
 {
     JSObject *function;
     JSObject *global;
@@ -1424,7 +1666,7 @@ function_new(JSContext      *context,
     }
 
     priv = priv_from_js(context, function);
-    if (!init_cached_function_data(context, priv, info))
+    if (!init_cached_function_data(context, priv, gtype, (GICallableInfo *)info))
       return NULL;
 
     return function;
@@ -1433,13 +1675,19 @@ function_new(JSContext      *context,
 JSObject*
 gjs_define_function(JSContext      *context,
                     JSObject       *in_object,
-                    GIFunctionInfo *info)
+                    GType           gtype,
+                    GICallableInfo *info)
 {
-    JSObject *function;
+    JSObject *function = NULL;
+    GIInfoType info_type;
+    gchar *name;
+    gboolean free_name;
+
+    info_type = g_base_info_get_type((GIBaseInfo *)info);
 
     JS_BeginRequest(context);
 
-    function = function_new(context, info);
+    function = function_new(context, gtype, info);
     if (function == NULL) {
         gjs_move_exception(context, context);
 
@@ -1447,8 +1695,17 @@ gjs_define_function(JSContext      *context,
         return NULL;
     }
 
-    if (!JS_DefineProperty(context, in_object,
-                           g_base_info_get_name( (GIBaseInfo*) info),
+    if (info_type == GI_INFO_TYPE_FUNCTION) {
+        name = (gchar *) g_base_info_get_name((GIBaseInfo*) info);
+        free_name = FALSE;
+    } else if (info_type == GI_INFO_TYPE_VFUNC) {
+        name = g_strdup_printf("vfunc_%s", g_base_info_get_name((GIBaseInfo*) info));
+        free_name = TRUE;
+    } else {
+        g_assert_not_reached ();
+    }
+
+    if (!JS_DefineProperty(context, in_object, name,
                            OBJECT_TO_JSVAL(function),
                            NULL, NULL,
                            GJS_MODULE_PROP_FLAGS)) {
@@ -1458,6 +1715,9 @@ gjs_define_function(JSContext      *context,
         return NULL;
     }
 
+    if (free_name)
+        g_free(name);
+
     JS_EndRequest(context);
     return function;
 }
@@ -1475,7 +1735,7 @@ gjs_invoke_c_function_uncached (JSContext      *context,
   JSBool result;
 
   memset (&function, 0, sizeof (Function));
-  if (!init_cached_function_data (context, &function, info))
+  if (!init_cached_function_data (context, &function, 0, info))
     return JS_FALSE;
 
   result = gjs_invoke_c_function (context, &function, obj, argc, argv, rval);
