@@ -27,6 +27,7 @@
 #include "ns.h"
 #include "function.h"
 #include "object.h"
+#include "param.h"
 #include "boxed.h"
 #include "union.h"
 #include "enumeration.h"
@@ -42,6 +43,8 @@
 #include <girepository.h>
 #include <string.h>
 
+#define DUMPBIN "_gjs_private"
+
 typedef struct {
     void *dummy;
 
@@ -322,6 +325,8 @@ repo_new(JSContext *context)
 
     g_assert(gjs_object_has_property(context, repo, "versions"));
 
+    JS_DefineObject(context, repo, DUMPBIN, NULL, NULL, JSPROP_PERMANENT);
+
     /* FIXME - hack to make namespaces load, since
      * gobject-introspection does not yet search a path properly.
      */
@@ -365,7 +370,7 @@ gjs_define_constant(JSContext      *context,
     type_info = g_constant_info_get_type(info);
     g_constant_info_get_value(info, &garg);
 
-    if (!gjs_value_from_g_argument(context, &value, type_info, &garg))
+    if (!gjs_value_from_g_argument(context, &value, type_info, &garg, TRUE))
         goto out;
 
     name = g_base_info_get_name((GIBaseInfo*) info);
@@ -460,7 +465,7 @@ gjs_define_info(JSContext  *context,
     case GI_INFO_TYPE_FUNCTION:
         {
             JSObject *f;
-            f = gjs_define_function(context, in_object, (GIFunctionInfo*) info);
+            f = gjs_define_function(context, in_object, 0, (GICallableInfo*) info);
             if (f == NULL)
                 return JS_FALSE;
         }
@@ -468,12 +473,20 @@ gjs_define_info(JSContext  *context,
     case GI_INFO_TYPE_OBJECT:
         {
             GType gtype;
-            GIBaseInfo *info_for_gtype;
             gtype = g_registered_type_info_get_g_type((GIRegisteredTypeInfo*)info);
-            if (!gjs_define_object_class(context, in_object, gtype, NULL, NULL, &info_for_gtype))
+
+            if (g_type_is_a (gtype, G_TYPE_PARAM)) {
+                if (!gjs_define_param_class(context, in_object, NULL))
+                    return JS_FALSE;
+            } else if (g_type_is_a (gtype, G_TYPE_OBJECT)) {
+                if (!gjs_define_object_class(context, in_object, gtype, NULL, NULL))
+                    return JS_FALSE;
+            } else {
+                gjs_throw (context,
+                           "Unsupported type %s, deriving from fundamental %s",
+                           g_type_name(gtype), g_type_name(g_type_fundamental(gtype)));
                 return JS_FALSE;
-            g_assert(g_base_info_equal(info, info_for_gtype));
-            g_base_info_unref(info_for_gtype);
+            }
         }
         break;
     case GI_INFO_TYPE_STRUCT:
@@ -509,6 +522,13 @@ gjs_define_info(JSContext  *context,
     return JS_TRUE;
 }
 
+/* Get the "unknown namespace", which should be used for unnamespaced types */
+JSObject*
+gjs_lookup_private_namespace(JSContext *context)
+{
+    return gjs_lookup_namespace_object_by_name(context, DUMPBIN);
+}
+
 /* Get the namespace object that the GIBaseInfo should be inside */
 JSObject*
 gjs_lookup_namespace_object(JSContext  *context,
